#!/bin/bash
#
# do_test.sh
#
#	splice from socket test program
#	Copyright (c) 2013 NETGEAR
#	Author: Hiro Sugawara

# This script tests "splice directly from scoket" by transferring file data
# over a TCP socket from the local host to a remove host-under-test.
# The remote host must run a tiny script that executes a command line
# send from the local host to remote hosts' TCP port 8000. Data is send to
# remote host's TCP port 8001
# Below is an exmple of the remote script:

#	#!/bin/sh
#	while true
#	do
#		cmd=`nc -l 8000`
#		[ "$cmd" ] || exit 0
#		$cmd
#	done

# Test strategy:
# Transfer data from local host to remote host to splice the data in
# a variety of ways:
#	Splice directly from socket
#	Splice-pipe-splice
#	Read from socket and write to file
#
#	Truncate/append to/overwrite middle of destination
#	Write across page boundaries or not
# Then compare the resulting files between different splice ways for checking
# integrity.

# PORT0 must agree with the remote script.
[ "$PORT0" ] || PORT0=8000	# Remote host listens on this port for commands
[ "$PORT1" ] || PORT1=8001	# Remote host receives data for splice
[ "$PORT2" ] || PORT2=10002	# Local host port to receive command output
HOSTNAME=`hostname`

# sps test may block because of pipe buffer size.
# x86_64 accepts 16K buffer size but ARM does not (passed with -b).

# Common additional arguments for splice_test
##SPLICE_TEST_ARGS="-b 12k"

MYNAME=`basename $0`
SOURCE=$1
DEST=$2

helpexit() {
	cat <<-_END >&2
$MYNAME: $1
Usage: $MYNAME FILE HOST:/DIR
  This script tests the direct splice-from-socket-to-file kernel feature for
  different file systems using the following transfer methods as references.
  - Conventional socket-to-user-write-to-file (rw)
  - Conventional socket-splice-to-pipe-splice-to-file (sps)

  Local FILE is sent over TCP to remote splice into HOST:/DIR/FILE.X.
  DIR must be the mount point of the file system under test.
  FILE may be an arbitrary file existing on the local host for exercising
  the target kernel.
  .X is the file name suffix appended according to the transfer menthod.
    rw:  read-write
    sps: splice-pipe-splice
    s:   direct splice from socket

  For details of various splice options, refer to 'splice_test --help'.
_END
	exit 1
}

[ "$DEST" ] || helpexit "No destination."
[ -f "$SOURCE" ] || helpexit "'$SOURCE' does not exist."
file=`basename $SOURCE`

HOST=`echo $DEST | sed 's|:.*$||'`
DIR=`echo $DEST | sed 's|^.*:||'`
[ -z "$HOST" -o -z "$DIR" -o "$HOST" = "$DIR" ] &&	\
	helpexit "Bad destination"

#
# Send a shell command line to remote host
#
send_cmd() {
	until echo "$@" | nc $HOST $PORT0
	do
		sleep 0.1
	done
}

#
# Send a shell command line to remote host and get its output
#
ask_cmd() {
	send_cmd "@$HOSTNAME:$PORT2 $1"
	nc -l $PORT2
}

[ $DIR = `ask_cmd "ls -d $DIR"` ] || helpexit "'$HOST:$DIR' does not exist."

if [ -z "$SPLICE_TEST" ]
then
	SPLICE_TEST=/tmp/splice_test
	case `ask_cmd "uname -m"` in
		arm*)
			SPLICE_TEST=${SPLICE_TEST}_arm	;;
	esac
fi
SPLICE_TEST="$SPLICE_TEST $SPLICE_TEST_ARGS"

MNT=`ask_cmd "mount | grep ' on $DIR type'"`

MNTDEV=`echo $MNT | cut -d ' ' -f 1`
MNTPNT=`echo $MNT | cut -d ' ' -f 3`
MNTTYP=`echo $MNT | cut -d ' ' -f 5`

[ "$MNTDEV" -a "$MNTPNT" -a "$MNTTYP" ] || helpexit "Unknown mount information"

echo $MNTDEV $MNTPNT

#
# Compare a spliced file against reference files
#
cmp_files() {
	local mode

	for mode in $1
	do
		[ $mode = s ] && continue
		send_cmd "echo Comparing: $DIR/$file.s $DIR/$file.$mode"
		send_cmd "cmp $DIR/$file.s $DIR/$file.$mode"
	done
}

mode2desc() {
	case "$1" in
		s)	echo "splice directly";;
		sps)	echo "splice-pipe-splice";;
		rw)	echo "read-write";;
	esac
}

#
# Perforn a test mode
#	syntax: exec_test nc_feeder splice_test_options mode
#
exec_mode() {
	local feeder="$1"
	local sopts="$2"
	local mode="$3"
	local desc=`mode2desc $mode`

	send_cmd "echo Testing: $feeder => splice $SPLICE_TEST_ARGS -m $mode $sopts ($desc)"
	send_cmd "$SPLICE_TEST -m $mode -i :$PORT1 -o $DIR/$file.$mode $sopts"
	sleep 1
	$feeder < $SOURCE 2>/dev/null | nc $HOST $PORT1
}

#
# Prform all (default) for given test modes
#	syntax: exec_test nc_feeder splice_test_options [mode ...]
#
exec_test() {
	local feeder="$1"; shift
	local sopts="$1"; shift
	local modes="s sps rw"

	if [ "$*" ]
	then
		modes="s $*"
	fi

	for mode in $modes
	do
		exec_mode "$feeder" "$sopts" $mode
	done

	cmp_files "$modes"

	send_cmd "echo Unmounting and re-mounting $MNTDEV at $MNTPNT"
	send_cmd "umount $MNTDEV"
	send_cmd "fsck $MNTDEV"
	send_cmd "mount -t $MNTTYP $MNTDEV $MNTPNT"

	cmp_files "$modes"
	send_cmd echo
}

banner() {
	send_cmd "echo === $1 ==="
}

######################################################################
#
# Script starts here.
#
######################################################################
send_cmd "echo =============== splice_from_socket test ==============="
send_cmd "date"
send_cmd "uname -a"
send_cmd "echo $HOST:$DIR"
send_cmd "df $DIR"
send_cmd "echo ======================================================="
send_cmd "echo Source file:"
send_cmd "echo `ls -l $SOURCE`"
send_cmd echo

#
# Tests start here
# Warning: Buffer size (-b) greater than 12K may block sps.
#

banner "Truncate destination"
exec_test "cat" "-t"

banner "Append to destination"
exec_test "cat" "-a"

banner "Overwrite first 100 bytes with default buffer size"
exec_test "dd bs=100 count=1"

banner "Overwrite first 100 bytes with exact buffer size"
exec_test "dd bs=100 count=1" "-b 100"

banner "Seek to 100 then overwrite 1000 bytes"
exec_test "dd bs=1000 count=1 skip=1" "-d 100"

banner "Seek to 4000 then overwrite 1000 bytes="
exec_test "dd bs=1000 count=1" "-d 4000"

banner "Seek to 4000000 then write 1000 bytes with 12K chunk size"
exec_test "dd bs=1000 count=1" "-d 4000000 -b 12k"

banner "One whole page write at page boudary"
exec_test "dd bs=4096 count=1" "-d 4096 -b 8k"

banner "Two whole page write at page boudary"
exec_test "dd bs=8192 count=1" "-d 8192 -b 8k"

banner "From the beginning of a page"
exec_test "dd bs=1000 count=1" "-d 16384 -b 12k"

banner "Toward the end of a page"
exec_test "dd bs=1000 count=1" "-d 15384 -b 8k"

banner "Zero write with time-out"
exec_test "sleep 10" "-a"

banner "Big buffer (sps excluded)"
exec_test "cat" "-d 20000 -b 64k" rw

send_cmd "df $DIR"
send_cmd ""
