/*
 * splice_test.c
 *
 * Simple test program for splice(2) and family
 *
 * Original code from ttp://d.hatena.ne.jp/viver/20071011/p1
 *	Copyright (c) 2007 FURUHASHI tagayuki
 *
 * Modified for mote generic use.
 *	Copyright (c) 2012-2013 Hiro Sugawara
 *	Copyright (c) 2012-2013 NETGEAR
 */
#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sendfile.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <errno.h>
#include <err.h>

/* splice(), vmsplice() */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#include <fcntl.h>
#undef _GNU_SOURCE
#else
#include <fcntl.h>
#endif
#include "splice_compat.h"

#ifndef PAGE_SIZE
#define PAGE_SIZE	getpagesize()
#endif
#ifndef PAGE_MASK
#define PAGE_MASK	(PAGE_SIZE - 1)
#endif

#ifndef BUFFER_SIZE
#define BUFFER_SIZE (16*1024)
#endif

size_t buffer_size = BUFFER_SIZE;
unsigned long repeat = LONG_MAX;

/* read()してwrite()する */
/* read() and write() */
size_t test_read_write(int infd, int outfd)
{
	char *buf = alloca(buffer_size);
	size_t total = 0;

	while (repeat-- > 0) {
		ssize_t written = 0, rlen = read(infd, buf, buffer_size);
		if (!rlen)
			break;
		if (rlen < 0)
			err(1, "%s: read()", __func__);
		do {
			ssize_t wlen = write(outfd, buf, rlen - written);

			if (wlen < 0)
				err(1, "%s: write()", __func__);
			written += wlen;
		} while (written < rlen);
		total += written;
	}
	return total;
}

/* 直接sendfile() */
/* Directly with sendfile() */
size_t test_sendfile(int infd, int outfd)
{
	ssize_t sent;
	size_t total = 0;

	while (repeat-- > 0 &&
		(sent = sendfile(outfd, infd, NULL, buffer_size)) > 0)
		total += sent;
	if (errno)
		err(1, "%s: ", __func__);
	return total;
}

/* 直接splice() */
/* Directly with splice() */
size_t test_splice(int infd, int outfd)
{
	ssize_t rcvd;
	size_t total = 0;

	while (repeat-- > 0 &&
		((rcvd = splice(infd, NULL, outfd, NULL, buffer_size, 0)) > 0)) 
		total += rcvd;
	if (errno)
		err(1, "%s: ", __func__);
	return total;
}

/* infdからパイプにsplice()して、パイプからoutfdにsplice()する */
/* splice() from infd to pipe, then splice() from pipe to outfd */
size_t test_splice_pipe_splice(int infd, int outfd)
{
	int pipedes[2];
	size_t total = 0;

	if (pipe(pipedes) < 0)
		err(1, "%s: pipe()", __func__);

	while (repeat-- > 0) {
		ssize_t written = 0;
		ssize_t rlen = splice(infd, NULL, pipedes[1], NULL, buffer_size,
				SPLICE_F_MOVE);
		if (!rlen)
			break;
		if (rlen < 0)
			err(1, "%s: splice() read", __func__);

		while (written < rlen) {
			ssize_t wlen = splice(pipedes[0], NULL, outfd, NULL,
					rlen - written, SPLICE_F_MOVE);
			if (wlen < 0)
				err(1, "%s: splice() write", __func__);
			written += wlen;
		}
		total += written;
	}
	return total;
}

/* infdからバッファにread()して、バッファからパイプにsplice()して、
 * パイプからoutfdにsplice()する */
/* read() from infd to buffer, splice() from buffer to pipe,
 * then splice() from pipe to outfd */
size_t test_read_vmsplice_copy_pipe_splice(int infd, int outfd)
{
	char *buf = alloca(buffer_size);
	int pipedes[2];
	size_t total = 0;

	if (pipe(pipedes) < 0)
		err(1, "%s: pipe()", __func__);

	while (repeat-- > 0) {
		ssize_t vm_written = 0, rlen;

		rlen = read(infd, buf, buffer_size);
		if (!rlen)
			break;
		if (rlen < 0)
			err(1, "%s: read()", __func__);

		while (vm_written < rlen) {
			ssize_t sp_written = 0;
			ssize_t vm_len;
			struct iovec vec;

			vec.iov_base = buf + vm_written;  /* 修正 */
			vec.iov_len = rlen - vm_written;
			vm_len = vmsplice(pipedes[1], &vec, 1, 0);
			/*vm_len = vmsplice(pipedes[1], &vec, 1, SPLICE_F_GIFT);*/
			if (vm_len < 0)
				err(1, "%s: vmsplice()", __func__);
			do {
				ssize_t sp_len = splice(pipedes[0], NULL,
						outfd, NULL,
						vm_len - sp_written,
						SPLICE_F_MOVE);
				if (sp_len < 0)
					err(1, "%s: splice()", __func__);
				sp_written += sp_len;
			} while (sp_written < vm_len);
			vm_written += vm_len;
		}
		total += vm_written;
	}
	return total;
}

/* infdからmmap()で確保したバッファ（ページ境界に合っている）にread()して、
 * バッファからパイプにsplice()して、パイプからoutfdにsplice()する */
/* read() from infd to page-aligned mmap'd buffer,
 * splice() from buffer to pipe, then splice() from pipe to outfd */
size_t test_mmap_read_vmsplice_gift_pipe_splice(int infd, int outfd)
{
	void *buf;
	int pipedes[2];
	ssize_t rlen, vm_len, sp_len;
	ssize_t vm_written, sp_written;
	struct iovec vec;
	ssize_t page_over;
	size_t total;

	if (pipe(pipedes) < 0)
		err(1, "%s: pipe()", __func__);

	/* ページ境界に合ったメモリを取得 */
	/* 単数切り上げでオーバーフローしないように多めに確保 */
	buf = mmap(NULL, buffer_size + PAGE_SIZE, PROT_READ | PROT_WRITE,
			MAP_PRIVATE | MAP_ANONYMOUS,
			-1, 0);
	if (!buf)
		err(1, "%s: mmap()", __func__);

	while (repeat-- > 0) {
		/* ここでreadするバッファサイズはページ境界に合っていた方が効率が良い */
		rlen = read(infd, buf, buffer_size);
		if (!rlen)
			break;
		if (rlen < 0)
			err(1, "%s: read()", __func__);

		vm_written = 0;
		while (vm_written < rlen) {
			vec.iov_base = buf + vm_written;  /* 修正 */
			/* 端数切り上げでページ境界に合わせる */
			vec.iov_len = (rlen - vm_written + PAGE_SIZE - 1) & PAGE_MASK;
			vm_len = vmsplice(pipedes[1], &vec, 1, SPLICE_F_GIFT);
			if (vm_len < 0)
				err(1, "%s: vmsplice()", __func__);
			/* 切り上げ分を元に戻す */
			if (vm_len > (rlen - vm_written)) {
				page_over = vm_len - (rlen - vm_written);
				vm_len = rlen - vm_written;
			} else {
				page_over = 0;
			}
			sp_written = 0;
			do {
				sp_len = splice(pipedes[0], NULL, outfd, NULL,
						vm_len - sp_written,
						SPLICE_F_MOVE);
				if (sp_len < 0)
					err(1, "%s: splice()", __func__);
				sp_written += sp_len;
			} while(sp_written < vm_len);
			/* pipeに残った分を捨てる（捨てないとバッファが一杯になってしまう） */
			if (page_over)
				read(pipedes[0], buf, page_over);
			vm_written += vm_len;
		}
		total += vm_written;
	}
	return total;
}


int infd_socket(uint16_t port)
{
	int lsock, infd;
	struct sockaddr_in addr;
	int on = 1;

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;
	lsock = socket(AF_INET, SOCK_STREAM, 0);
	if (lsock < 0)
		err(1, "%s: socket()", __func__);
	if (setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
		err(1, "%s: socketopt()", __func__);
	if (bind(lsock, (struct sockaddr*)&addr, sizeof(addr)) < 0 )
		err(1, "%s: bind()", __func__);
	if (listen(lsock, 1) < 0)
		err(1, "%s: listen()", __func__);


	printf("accept...\n");
	infd = accept(lsock, NULL, NULL);
	if (infd < 0)
		err(1, "%s: accept()", __func__);
	printf("accept ok\n");

	close(lsock);

	return infd;
}

int outfd_socket(uint16_t port, const char *host)
{
	int outfd;
	struct sockaddr_in addr;

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	outfd = socket(AF_INET, SOCK_STREAM, 0);
	if (outfd < 0)
		err(1, "%s: socket()", __func__);
	if (!inet_aton(host, &addr.sin_addr) )
		err(1, "%s: inet_aton()", __func__);
	if (connect(outfd, (struct sockaddr*)&addr, sizeof(addr)) < 0)
		err(1, "%s: connect()", __func__);

	return outfd;
}


int infd_file(const char *file)
{
	int infd = !strcmp(file, "-") ? dup(0) : open(file, O_RDONLY);
	if (infd < 0)
		err(1, "%s: open/dup()", __func__);
	return infd;
}

int outfd_file(const char *file, int o_mode)
{
	int outfd = !strcmp(file, "-")	? dup(1)
					: open(file, O_WRONLY | O_CREAT, 0644);
	if (outfd < 0)
		err(1, "%s: open/dup()", __func__);
	if (o_mode & O_TRUNC)
		ftruncate(outfd, 0);
	if (o_mode & O_APPEND)
		lseek(outfd, 0, SEEK_END);
	return outfd;
}

int infd_open(char *in)
{
	char *p = strchr(in, ':');

	if (p) {
		int port;

		*p++ = '\0';
		port = atoi(p);
		if (port <= 0)
			errx(1, "Bad port - in");
		return infd_socket(port);
	}
	return infd_file(in);
}

int outfd_open(char *out, int o_mode)
{
	char *p = strchr(out, ':');

	if (p) {
		int port;

		*p++ = '\0';
		port = atoi(p);
		if (port <= 0)
			errx(1, "Bad port - out");
		return outfd_socket(port, out);
	}
	return outfd_file(out, o_mode);
}

void helpexit(int code)
{
	errx(code, "\b\b [opts]\n"
"  -a                  Tppend to ouput\n"
"  -b <buffer_size>    (default=%d)\n"
"  -d <output_seek_to> (octal, decimal, hexadecimal)\n"
"  -i <input>\n"
"  -m <mode>\n"
"  -o <output>\n"
"  -r <repeat_count>\n"
"  -s <input_seek_to>  (octal, decimal, hexadecimal)\n"
"  -t [<size>]         Truncate output\n"
"\n"
"     input/output\n"
"       file or - or addr:port (TCP only)\n"
"\n"
"     mode\n"
"       rw      read/write\n"
"       sf      sendfile\n"
"       s       splice\n"
"       sps     splice/pipe/splice\n"
"       rvcps   read/vmsplice/copy/pipe/splice\n"
"       mrvgps  mmap/read/vmsplice/gift/pipe/splice\n"
		, BUFFER_SIZE
	);
}

int main(int argc, char *argv[])
{
	char *inf = NULL, *outf = NULL;
	int infd, outfd;
	size_t (*tester)(int, int) = NULL;
	int opt;
	int o_mode = 0;
	off64_t *pos = NULL, spos = -1, dpos = -1;
	size_t xfrd;
	struct timeval start_tv, end_tv;
	time_t sec;
	suseconds_t usec;

	while ((opt = getopt(argc, argv, "ab:d:hi:m:o:r:s:t")) != -1) {
		char c;

		switch (opt) {
		case 'a':
			if (dpos >= 0 || (o_mode & O_TRUNC))
				errx(1, "-a, -d and -t are exclusive");
			o_mode |= O_APPEND;
			break;
		case 'b':
			switch (sscanf(optarg, "%lu%c", &buffer_size, &c)) {
			case 2:
				switch (c) {
				case 'k':
				case 'K':
					buffer_size *= 1024;
					break;
				case 'M':
					buffer_size *= 1024 * 1024;
					break;
				default:
					helpexit(1);
				}
			case 1:
				break;
			default:
				helpexit(1);
			}
			break;
		case 'h':
			helpexit(0);
			break;
		case 'm':
			if (tester)
				errx(1, "Choose only one test mode");
			if (!strcmp(optarg, "rw"))
				tester = test_read_write;
			else if (!strcmp(optarg, "sf"))
				tester = test_sendfile;
			else if (!strcmp(optarg, "s"))
				tester = test_splice;
			else if (!strcmp(optarg, "sps"))
				tester = test_splice_pipe_splice;
			else if (!strcmp(optarg, "rvcps"))
				tester = test_read_vmsplice_copy_pipe_splice;
			else if (!strcmp(optarg, "mrvgps"))
				tester =
				test_mmap_read_vmsplice_gift_pipe_splice;
			else
				helpexit(1);
			break;
		case 'i':
			if (inf)
				errx(1, "Only one input");
			inf = optarg;
			break;
		case 'o':
			if (outf)
				errx(1, "Only one output");
			outf = optarg;
			break;
		case 'r':
			repeat = strtoul(optarg, NULL, 0);
			if (errno)
				errx(1, "Bad repeat count: %s", optarg);
			break;
		case 'd':
			if (o_mode & (O_APPEND | O_TRUNC))
				errx(1, "-a, -d and -t are exclusive");
			pos = &dpos;
			/* Falls through */
		case 's':
			if (!pos)
				pos = &spos;
			errno = 0;
			*pos = strtoull(optarg, NULL, 0);
			if (errno)
				errx(1, "Bad seek position: %s", optarg);
			pos = NULL;
			break;
		case 't':
			if (dpos >= 0 || (o_mode & O_APPEND))
				errx(1, "-a, -d and -t are exclusive");
			o_mode |= O_TRUNC;
			break;
		default:
			helpexit(1);
		}
	}

	if (optind < argc)
		errx(1, "Extra argument(s) after options");
	if (!tester)
		errx(1, "Choose one test mode");


	infd = infd_open(inf);
	outfd = outfd_open(outf, o_mode);
	if (infd < 0 || outfd < 0)
		errx(1, "Either -i or-o is missing or bad");
	if (dpos > 0)
		lseek64(outfd, dpos, SEEK_SET);
	if (spos > 0)
		lseek64(infd, spos, SEEK_SET);

	gettimeofday(&start_tv, NULL);
	xfrd = tester(infd, outfd);
	gettimeofday(&end_tv, NULL);
	sec	= end_tv.tv_sec - start_tv.tv_sec;
	usec	= end_tv.tv_usec - start_tv.tv_usec;
	if (usec < 0) {
		sec--;
		usec += 1000000;
	}

	printf("Transferred: %lu bytes in %lu.%06ld seconds (%.0fkB/s)\n",
		xfrd,
		sec, usec,
		(double)xfrd/((double)sec * 1000 + (double)usec/1000)
		);
	return 0;
}

